//
// GameX (Release 2 - Vectors / Matricies / CameraX)
// CameraX Class Header
// 
// Copyright (C) 2002 Rama C. Hoetzlein
// 
// This software is released under the GameX GNU GPL 
// Open Source License. See the GameX documentation included
// with this source code for terms of modification, 
// distribution and re-release. 
//

//#include "gamex-win-dx.h"


#ifndef GAMEX_CAMERA_H_INCLUDED
#define GAMEX_CAMERA_H_INCLUDED

#include "gamex-global-def.h"


///	#include "gamex-matrix.h"
///	#include "gamex-vector.h"

#ifndef DEGtoRAD
	#define DEGtoRAD (3.1415926538f/180.0f)
#endif

#define CAMERA_PERSPECT		0
#define CAMERA_PARALLEL		1

#include "gamex-matrix.h"

class CameraX
{
public:
	CameraX ();

	void Project (const Vec3f &p, float &x, float &y, float &z); // projects a point from 3D to 2D
	void ProjectFast (const Vec3f &p, float &x, float &y, float &z); // same as Project() but with less checks and no support for SetOffset()

	void SetFromPosition (const Vec3f &from);
	void SetFromAngles (const Vec3f &angs); // x is x rotation (degrees), y is pitch, z is zoom distance
	void SetToPosition (const Vec3f &to);
	void SetUpVector (const Vec3f &up);
	void SetNearPlane (float n);
	void SetRollAngle (float angle);
	inline void SetOffset (float x, float y) {z_offset_x = x; z_offset_y = y;}
	void SetFarPlane (float f);
	void SetViewHeight (float h);
	void SetProjection (int proj_const);
	inline void SetNearClip (float zdepth) {z_near_clip = zdepth;}
	inline float GetNearClip () {return z_near_clip;}

	void SetWindow (int x1, int y1, int x2, int y2);
	void GetWindow (int&x1, int&y1, int&x2, int&y2);

	void ViewTransform (Vec3f &p, Vec4f &q);
	void ProjectTransform (Vec4f &q);
	void ScreenTransform (Vec4f &q, Vec3f &p);
	void CameraTransform (Vec3f &p, Vec4f &q);
	void WindowTransform (Vec4f &q, float &x, float &y, float &z);
	void InverseTransform (Vec3f &p, Vec4f &q);
	
	inline float GetRollAngle () const {return z_roll;}
	inline float GetHeight () const {return z_height_dist;}
	inline float GetNear () const {return z_near_dist;}
	inline float GetFar () const {return z_far_dist;}
	inline float GetInvPlaneSep () const {return z_plane_sep;}
	inline float GetPixelsPerUnit() const {return (z_cam_x2-z_cam_x1);}
	inline const Vec3f &From () const {return z_from_pos;}
	inline const Vec3f &GetFromPosition () const {return z_from_pos;}
	inline const Vec3f &To () const {return z_to_pos;}
	inline const Vec3f &GetToPosition () const {return z_to_pos;}
	inline const Mat44f * GetViewMatrix() const {if(z_view_matrix_invalid) z_MakeViewMatrix(); return &z_view_matrix;}
	inline const Mat44f * GetProjMatrix() const {if(z_proj_matrix_invalid) z_MakeProjMatrix(); return &z_proj_matrix;}
	inline const Vec3f & GetDirVector() const {if(z_dir_vector_invalid) z_MakeDirVector(); return z_dir_vector;}
	inline const Vec3f & GetUpVector() const {return z_up_vector;}
	inline const Vec3f & GetSideVector() const {return z_side_vector;}
private:
	void z_MakeDirVector () const;			// Calculate direction (using to and from vectors)
	void z_MakeViewMatrix () const;			// Calculate view matrix (using dir, from and side vectors)
	void z_MakeProjMatrix () const;			// Calculate projection matrix (using proj. parameters)
	void z_MakeTransformMatrix () const;

	int z_proj_type;					// Camera projection type

	Vec3f z_from_pos, z_to_pos;		// Camera To and From positions
	float z_roll;						// Camera twist about its axis
									// Camera Tview vectors (N, V, and U)
	float z_offset_x, z_offset_y;
	mutable Vec3f z_dir_vector, z_up_vector, z_side_vector;
	mutable Mat44f z_view_matrix;			// Camera Tview matrix
									// Camera Tpersp parameters (d, f, and h)		
	float z_near_dist, z_far_dist, z_height_dist, z_near_clip, z_plane_sep;	
	mutable Mat44f z_proj_matrix;			// Camera Tpersp (also Tparallel or Tortho)

	mutable Mat44f z_transform_matrix;		// Camera Transform (Tview * Tpersp)

	mutable Mat44f z_basis, z_rollM;
	mutable Mat44f z_inv_proj, z_inv_view;

	float z_cam_x1, z_cam_y1;				// Camera Window (Maps -1<x,y<1 to 0<x,y<win)
	float z_cam_x2, z_cam_y2;
	float z_cam_sx, z_cam_sy;
	float z_cam_cx, z_cam_cy;

	mutable bool z_dir_vector_invalid, z_view_matrix_invalid, z_proj_matrix_invalid, z_transform_matrix_invalid; // dirty variable flags
};

#endif