//
// GameX - File Class Code 
// 
// Copyright (C) 2002 Rama C. Hoetzlein
// 
// This software is released under the GameX GNU GPL 
// Open Source License. See the GameX documentation included
// with this source code for terms of modification, 
// distribution and re-release. 
// 

#include "gamex-file.h"

File::File ()
{
	filemodes = FILE_NOTOPEN;
	#ifdef FILE_COUNT_WRITES
		write_number = 0;
	#endif
}


File::File (char *fname, FileFlags fmodes)
{
	Open (fname, fmodes);
}

File::~File ()
{
	Close ();	
}

int File::Exist (char *fname)
{
 	FILE *f = fopen (fname, "r");
	if(f == NULL)
	{
		return FILE_EXIST_NO;
	}
	else
	{
		fclose (f);
		return FILE_EXIST_YES;
	}
}

int File::Open (char *fname, FileFlags fmodes)
{
	char mode[10];
	
	// Initialize translation between Visual C++ fopen modes and File class open modes
	strcpy (mode, "\0");
	
	// Handle FILE_APPEND, FILE_READ, FILE_WRITE, FILE_AUTOCREATE, and FILE_LEAVEEOF
	if((fmodes & FILE_APPEND) == FILE_APPEND)
	{
		if((fmodes & FILE_READ) == FILE_READ)
		{
			#ifdef FILE_DEBUG
				debug.Output ("File::Open: Cannot open file in both APPEND and READ modes.\n");
			#endif
			return FILE_STATUS_READERROR;
		}
		if((fmodes & FILE_WRITE) == FILE_WRITE)
		{
			#ifdef FILE_DEBUG
				debug.Output ("File::Open: Cannot open file in both APPEND and WRITE modes.\n");
			#endif
			return FILE_STATUS_WRITEERROR;
		}
		strcat (mode, "a");

		// If FILE_AUTOCREATE is off, make sure file exists before appending
		if((fmodes & FILE_AUTOCREATE) != FILE_AUTOCREATE)
		{
			if(Exist(fname) == FILE_EXIST_NO)
			{
				#ifdef FILE_DEBUG
					debug.Output ("File::Open: File does not exist for append. (Use FILE_AUTOCREATE if desired)\n");
				#endif				
				return FILE_EXIST_NO;
			}
		}
		// If FILE_LEAVEEOF is off, default to remove EOF mode
		if((fmodes & FILE_LEAVEEOF) != FILE_LEAVEEOF)
			strcat (mode, "+");
	} else
	{
		if((fmodes & FILE_LEAVEEOF) == FILE_LEAVEEOF)
			#ifdef FILE_DEBUG
				debug.Output ("File::Open: (Warning only) FILE_LEAVEEOF only has meaning with FILE_APPEND.\n");
			#endif
		if((fmodes & FILE_READ) == FILE_READ) {			
			if((fmodes & FILE_WRITE) == FILE_WRITE) {				
				// If FILE_AUTOCREATE is on, autocreate file (otherwise, check for file existence)
				if((fmodes & FILE_AUTOCREATE) == FILE_AUTOCREATE)
				{
					strcat (mode, "w+");
				} else
				{
					if(Exist(fname) == FILE_EXIST_NO)
					{
						#ifdef FILE_DEBUG
							debug.Output ("File::Open: File does not exist for read/write. (Use FILE_AUTOCREATE if desired)\n");
						#endif
						return FILE_EXIST_NO;
					}
					strcat (mode, "r+");
				}
			} else
			{
				if((fmodes & FILE_AUTOCREATE) == FILE_AUTOCREATE)
				{
					#ifdef FILE_DEBUG
						debug.Output ("File::Open: (Warning only) FILE_AUTOCREATE has no meaning when using FILE_READ without FILE_APPEND or FILE_WRITE.\n");
					#endif
				}
				if(Exist(fname) == FILE_EXIST_NO)
				{
					#ifdef FILE_DEBUG
						debug.Output ("File::Open: File does not exist for read.\n");				
					#endif
					return FILE_EXIST_NO;
				}
				strcat (mode, "r");
			}
		} else if((fmodes & FILE_WRITE) == FILE_WRITE)
		{
			if((fmodes & FILE_AUTOCREATE) != FILE_AUTOCREATE)
			{
				if(Exist(fname) == FILE_EXIST_NO)
				{
					#ifdef FILE_DEBUG
						debug.Output ("File::Open: File does not exist for write (Use FILE_AUTOCREATE is desired).\n");						
					#endif
					return FILE_EXIST_NO;
				}
			}
			strcat (mode, "w");			
		} else
		{
			#ifdef FILE_DEBUG
				debug.Output ("File::Open: Must open file using FILE_APPEND, FILE_READ, or FILE_WRITE.\n");
			#endif
			return FILE_STATUS_READERROR;
		}
	}
	if((fmodes & FILE_NOTOPEN) == FILE_NOTOPEN)
	{
		#ifdef FILE_DEBUG
			debug.Output ("File::Open: Cannot open a file using FILE_NOTOPEN command!!\n");
		#endif
		fmodes ^= FILE_NOTOPEN;
	}

	// Handle FILE_RANDOM and FILE_SEQUENTIAL (mutually exclusive)	
	if((fmodes & FILE_RANDOM) == FILE_RANDOM)
	{
		if((fmodes & FILE_SEQUENTIAL) == FILE_SEQUENTIAL)
		{
			#ifdef FILE_DEBUG
				debug.Output ("File::Open: Cannot open in SEQUENTIAL and RANDOM access modes.\n");
			#endif	
			return FILE_STATUS_READERROR;
		} else
		{
			strcat (mode, "b");
		}		
	} else if((fmodes & FILE_SEQUENTIAL) == FILE_SEQUENTIAL)
	{
		if((fmodes & FILE_RAWDATA) == FILE_RAWDATA)
		{
			strcat (mode, "b");
		} else
		{
			strcat (mode, "t"); // can screw up proper loading of sequential binary data by counting certain values as end-of-file, even mid-file
		}
	} else
	{
		#ifdef FILE_DEBUG
			debug.Output ("File::Open: Must open in SEQUENTIAL or RANDOM access mode.\n");
		#endif		
		return FILE_STATUS_READERROR;
	}	
	
	// Open file
	strcpy (filename, fname);
	filemodes = fmodes;
	filehandle = fopen (filename, mode);
	if(filehandle == NULL) 
	{
		// happens if write access was requested and the file is marked as read-only
		filemodes |= FILE_NOTOPEN; // prevent error on close
		return FILE_STATUS_WRITEERROR; // return error so caller can react
	}
	filepos = 0;	
	return FILE_STATUS_OK;
}

void File::Close ()
{
	
	if((filemodes & FILE_NOTOPEN) != FILE_NOTOPEN)
	{
		fflush (filehandle);
		fclose (filehandle);
	}
	filemodes = FILE_NOTOPEN;
}

void File::Delete (char *fname)
{
	if(Exist(fname)) remove(fname);
}

int File::ReadC (int num, char *buf)
{
	size_t status = fread (buf, num, 1, filehandle);
	#ifdef FILE_CHECKS
		if((filemodes & FILE_READ) != FILE_READ) return FILE_STATUS_NOREAD;
		if(status < 1)
		{
			if(feof(filehandle) != 0) return FILE_STATUS_EOF;
			return FILE_STATUS_READERROR;
		}
	#endif
	return FILE_STATUS_OK;	
}

int File::WriteC (int num, char *buf)
{
	size_t status;
	status = fwrite (buf, num, 1, filehandle);
	#ifdef FILE_CHECKS
		if((filemodes & FILE_WRITE) != FILE_WRITE) return FILE_STATUS_NOWRITE;
		if(status < 1) return FILE_STATUS_WRITEERROR;
	#endif
	#ifdef FILE_COUNT_WRITES
		write_number += num;
	#endif
	return FILE_STATUS_OK;	
}

int File::WriteS (char * str)
{
	int n = 0;
	while(str[n++]);
	return WriteC(n-1,str);
}
int File::Read (int num, unsigned char *buf)
{
	buf[0] = 0;
	size_t status = fread (buf, 1, num, filehandle);
	#ifdef FILE_CHECKS
		if((filemodes & FILE_READ) != FILE_READ) return FILE_STATUS_NOREAD;
		if(status < 1)
		{
			if(feof(filehandle) != 0)
				return FILE_STATUS_EOF;
			return FILE_STATUS_READERROR;
		}
	#endif
	return FILE_STATUS_OK;	
}

int File::Write (int num, unsigned char *buf)
{
	size_t status;
	status = fwrite (buf, 1, num, filehandle);
	#ifdef FILE_CHECKS
		if((filemodes & FILE_WRITE) != FILE_WRITE) return FILE_STATUS_NOWRITE;
		if(status < 1) return FILE_STATUS_WRITEERROR;
	#endif
	#ifdef FILE_COUNT_WRITES
		write_number += num;
	#endif
	return FILE_STATUS_OK;	
}


int File::Read (unsigned char &c)
{
	c = (unsigned char) fgetc (filehandle);
	#ifdef FILE_CHECKS
		if((filemodes & FILE_READ) != FILE_READ) return FILE_STATUS_NOREAD;
		if(feof(filehandle) != 0)
			return FILE_STATUS_EOF;
		if(ferror(filehandle) != 0) return FILE_STATUS_READERROR;
	#endif 		
	return FILE_STATUS_OK;
}

int File::Write (unsigned char c)
{
	fputc (c, filehandle);
	#ifdef FILE_CHECKS	
		if((filemodes & FILE_WRITE) != FILE_WRITE) return FILE_STATUS_NOWRITE;
		if(ferror(filehandle) != 0) return FILE_STATUS_WRITEERROR;		
	#endif
	#ifdef FILE_COUNT_WRITES
		write_number ++;
	#endif
	return FILE_STATUS_OK;	
}

int File::Read2 (unsigned short &c)
{
///	// !!! For now, Read2 uses the Read function above 
///	// !!! This could be much faster if we read individual bytes directly
///	return Read (2, (unsigned char *) &c);

	unsigned char* buf = (unsigned char *) &c;
	buf[0] = 0;
	size_t status = fread (buf, 2, 1, filehandle);
	#ifdef FILE_CHECKS
		if((filemodes & FILE_READ) != FILE_READ) return FILE_STATUS_NOREAD;
		if(status < 1)
		{
			if(feof(filehandle) != 0)
				return FILE_STATUS_EOF;
			return FILE_STATUS_READERROR;
		}
	#endif
	return FILE_STATUS_OK;	
}

int File::Write2 (unsigned short c)
{
///	// !!! For now, Write2 uses the Write function above 
///	// !!! This could be much faster if we read individual bytes directly
///	return Write (2, (unsigned char *) &c);

	unsigned char* buf = (unsigned char *) &c;
	size_t status;
	status = fwrite (buf, 2, 1, filehandle);
	#ifdef FILE_CHECKS
		if((filemodes & FILE_WRITE) != FILE_WRITE) return FILE_STATUS_NOWRITE;
		if(status < 1) return FILE_STATUS_WRITEERROR;
	#endif
	#ifdef FILE_COUNT_WRITES
		write_number += num;
	#endif
	return FILE_STATUS_OK;	
}

int File::Read4 (unsigned long &c)
{
///	// !!! For now, Read4 uses the Read function above 
///	// !!! This could be much faster if we read individual bytes directly
///	return Read (4, (unsigned char *) &c);

	unsigned char* buf = (unsigned char *) &c;
	buf[0] = 0;
	size_t status = fread (buf, 4, 1, filehandle);
	#ifdef FILE_CHECKS
		if((filemodes & FILE_READ) != FILE_READ) return FILE_STATUS_NOREAD;
		if(status < 1)
		{
			if(feof(filehandle) != 0)
				return FILE_STATUS_EOF;
			return FILE_STATUS_READERROR;
		}
	#endif
	return FILE_STATUS_OK;	
}

int File::Write4 (unsigned long c)
{
///	// !!! For now, Write4 uses the Write function above 
///	// !!! This could be much faster if we read individual bytes directly
///	return Write (4, (unsigned char *) &c);

	unsigned char* buf = (unsigned char *) &c;
	size_t status;
	status = fwrite (buf, 4, 1, filehandle);
	#ifdef FILE_CHECKS
		if((filemodes & FILE_WRITE) != FILE_WRITE) return FILE_STATUS_NOWRITE;
		if(status < 1) return FILE_STATUS_WRITEERROR;
	#endif
	#ifdef FILE_COUNT_WRITES
		write_number += num;
	#endif
	return FILE_STATUS_OK;	
}

int File::ReadLine (char *buf)
{
	fgets (buf, FILE_LINELEN, filehandle);
	if(feof(filehandle) != 0) return FILE_STATUS_EOF;
	return FILE_STATUS_OK;
}

int File::WriteLine (char *buf)
{	
	fputs (buf, filehandle);
	return FILE_STATUS_OK;
}

void File::SetPosition (long pos)
{
	if(pos == FILE_POS_BEGIN)
	{
		fseek (filehandle, 0, SEEK_SET);
	} else if(pos == FILE_POS_END)
	{
		fseek (filehandle, 0, SEEK_END);
	} else
	{
		fseek (filehandle, pos, SEEK_SET);
	}
}

long File::GetPosition ()
{
	return ftell (filehandle);
}
