//
// GameX)
// Matrix Class Header
// 


#ifndef GAMEX_MATRIX_H_INCLUDED
#define GAMEX_MATRIX_H_INCLUDED

#include "gamex-global-def.h"
#include <math.h>
#define MT template <typename type, int rows, int cols>
#define MT2 template <typename type2, int rows2, int cols2>
#define MTM template <typename type2, int cols2>
#define MTD template <typename type, int rows>
#define T template <typename type>
#define T2 template <typename type2>
#define AN template <typename anything>
T struct Vec2; T struct Vec3; T struct Vec4;
MT struct Matrix;

// define some of the most common vector types
// (you can, of course, define other vector types similarly)
typedef Matrix<float,4,4> Mat44f; // 4x4 matrix of floats
typedef Matrix<float,3,4> Mat34f; // 3x4 matrix of floats
typedef Matrix<float,4,3> Mat43f; // 4x3 matrix of floats
typedef Matrix<float,3,3> Mat33f; // 3x3 matrix of floats
typedef Matrix<float,2,2> Mat22f; // 2x2 matrix of floats
typedef Matrix<long,4,4>  Mat44i; // 4x4 matrix of int
typedef Matrix<long,3,3>  Mat33i; // 3x3 matrix of int
typedef Matrix<long,2,2>  Mat22i; // 2x2 matrix of int
typedef Matrix<char,4,4>  Mat44c; // 4x4 matrix of chars

// alternate naming scheme for backward compatibility
typedef Mat44f MatrixF;
typedef Mat44i MatrixI;

// disable "possible loss of data" warnings since int times float can happen a lot naturally depending on which template types are used
#pragma warning(push)
#pragma warning(disable: 4244)

#pragma pack(push)
#pragma pack(1)

#define FORALL(something) FOR(r,Rows) FOR(c,Cols) something

// class definition
MT struct Matrix
{
	type data [rows][cols];

	const static int Rows = rows;
	const static int Cols = cols;
	inline int GetRows() const { return rows; }
	inline int GetCols() const { return cols; }

	// common matrices (example usage: Mat44f m = Mat44f::IdentityMatrix;)
	const static Matrix<type,rows,cols> ZeroMatrix;
	const static Matrix<type,rows,cols> IdentityMatrix;

	// constructors
MT2	inline Matrix<type,rows,cols> (const Matrix<type2,rows2,cols2>& m) {*this = m;}
T2	inline Matrix<type,rows,cols> (const type2 nArray [rows][cols]) {*this = nArray;}
T2	inline Matrix<type,rows,cols> (const type2* nArray) {*this = nArray;}
T2	inline Matrix<type,rows,cols> (const type2& n) {*this = n;}
	inline Matrix<type,rows,cols> () {*this =(type)0;}

	// operator overloads
T2	inline void operator = (const Matrix<type2,rows,cols>& op) { FORALL(data[r][c] = (type)op[r][c]); }
T2	inline void operator = (const type2 opArray [rows][cols]) { FORALL(data[r][c] = (type)opArray[r][c]); }
T2	inline void operator = (const type2* opArray) { FORALL(data[r][c] = (type)opArray[c+r*cols]); }
T2	inline void operator = (const type2& op) { FORALL(data[r][c] = (type)op); }

	// matrix multiplication - use Scale() if you want to do term-by-term multiplication instead
T2	void operator *= (const Matrix<type2,rows,cols>& op); 
MTM	Matrix<type,rows,cols2> operator * (const Matrix<type2,cols,cols2>& op) const;

T2	inline void operator += (const Matrix<type2,rows,cols>& op) { FORALL(data[r][c] = (type)(data[r][c] + op[r][c])); }
T2	inline void operator -= (const Matrix<type2,rows,cols>& op) { FORALL(data[r][c] = (type)(data[r][c] - op[r][c])); }
T2	inline void ScaleBy     (const Matrix<type2,rows,cols>& op) { FORALL(data[r][c] = (type)(data[r][c] * op[r][c])); }
T2	inline void operator /= (const Matrix<type2,rows,cols>& op) { FORALL(data[r][c] = (type)(data[r][c] / op[r][c])); }

// invalid operations - if you get here while debugging in VS.NET, press F10 to view your code
MT2	inline void operator = (const Matrix<type2,rows2,cols2>& op) { assertMsg(0,"Can't assign matrix to a matrix with different number of rows or columns."); }
MT2	inline Matrix<type,rows,cols2> operator * (const Matrix<type2,rows2,cols2>& op) const { assertMsg(0,"Can't multiply matrices together unless the number of rows in the first matrix equals the number of columns in the second matrix."); }
MT2	inline void operator += (const Matrix<type2,rows2,cols2>& op) { assertMsg(0,"Can't use += or + with two matrices that have a different number of rows or columns."); }
MT2	inline void operator -= (const Matrix<type2,rows2,cols2>& op) { assertMsg(0,"Can't use -= or * with two matrices that have a different number of rows or columns."); }
MT2	inline void ScaleBy     (const Matrix<type2,rows2,cols2>& op) { assertMsg(0,"Can't use ScaleBy() or Scale() with two matrices that have a different number of rows or columns."); }
MT2	inline void operator /= (const Matrix<type2,rows2,cols2>& op) { assertMsg(0,"Can't use /= or / with two matrices that have a different number of rows or columns."); }

T2	inline void operator += (const type2& op) { FORALL(data[r][c] = (type)(data[r][c] + op)); }
T2	inline void operator -= (const type2& op) { FORALL(data[r][c] = (type)(data[r][c] - op)); }
T2	inline void operator *= (const type2& op) { FORALL(data[r][c] = (type)(data[r][c] * op)); }
T2	inline void ScaleBy     (const type2& op) { FORALL(data[r][c] = (type)(data[r][c] * op)); }
T2	inline void operator /= (const type2& op) { FORALL(data[r][c] = (type)(data[r][c] / op)); }


AN	inline Matrix<type,rows,cols> operator + (const anything& op) const { Matrix<type,rows,cols> m; m += op; return m; }
AN	inline Matrix<type,rows,cols> operator - (const anything& op) const { Matrix<type,rows,cols> m; m -= op; return m; }
AN	inline Matrix<type,rows,cols> operator * (const anything& op) const { Matrix<type,rows,cols> m; m *= op; return m; }
AN	inline Matrix<type,rows,cols> Scale      (const anything& op) const { Matrix<type,rows,cols> m; m *= op; return m; }
AN	inline Matrix<type,rows,cols> operator / (const anything& op) const { Matrix<type,rows,cols> m; m /= op; return m; }

	


	// accessors ( NOTE: matrix(r,c) and matrix[r][c] should both work the same )
	inline type& operator () (int r, int c) { return data[r][c]; }
	inline const type& operator () (int r, int c) const { return data[r][c]; }
	inline type* operator [] (int r) { return data[r]; }
	inline const type* operator [] (int r) const { return data[r]; }
	inline type* GetData() { return (type*)(*data); }
	inline const type* GetData() const { return (const type*)(*data); }
	inline operator type** () { return (type**)(data); }
	inline operator const type** () const { return (const type**)(data); }

	void Transpose ();
	Matrix<type,cols,rows> GetTranspose () const;

	// functions specific to 4x4 matrices:
T2	static Matrix<type,4,4> BasisMatrix(const Vec3<type2> &c1, const Vec3<type2> &c2, const Vec3<type2> &c3)
	{
		Matrix<type,4,4> m;

		type *n = m.GetData();	
		*n++ = (type) c1.x; *n++ = (type) c2.x; *n++ = (type) c3.x; *n++ = (type) 0;
		*n++ = (type) c1.y; *n++ = (type) c2.y; *n++ = (type) c3.y; *n++ = (type) 0;
		*n++ = (type) c1.z; *n++ = (type) c2.z; *n++ = (type) c3.z; *n++ = (type) 0;
		*n++ = (type) 0; *n++ = (type) 0; *n++ = (type) 0; *n++ = (type) 0;

		return m;
	}

	static Matrix<type,4,4> RotateZMatrix(type sinAng, type cosAng)
	{
		Matrix<type,4,4> m;

		type* n = m.GetData();
		*n++ = cosAng;
		*n = -sinAng; n += 3;
		*n++ = sinAng;
		*n = cosAng; n += 5; 
		*n = 1; n += 5; *n = 1;

		return m;
	}
	static Matrix<type,4,4> RotateYMatrix(type sinAng, type cosAng)
	{
		Matrix<type,4,4> m;
		type *n = m.GetData();

		*n = cosAng; n += 2;
		*n = sinAng; n += 3;
		*n = 1; n += 3;
		*n = -sinAng; n += 2;
		*n = cosAng; n += 5;
		*n = 1;

		return m;
	}
	static Matrix<type,4,4> RotateXMatrix(type sinAng, type cosAng)
	{
		Matrix<type,4,4> m;
		type *n = m.GetData();

		*n = 1; n += 5;
		*n++ = cosAng;
		*n = -sinAng; n += 3;
		*n++ = sinAng;
		*n = cosAng; n += 5;
		*n = 1; 

		return m;
	}

	// in-place assignment (must be called on a 4x4 matrix or it won't compile) for some degree of backward compatibility
	inline void RotateZ (type sinAng, type cosAng) { *this = RotateZMatrix(sinAng,cosAng); }
	inline void RotateY (type sinAng, type cosAng) { *this = RotateYMatrix(sinAng,cosAng); }
	inline void RotateX (type sinAng, type cosAng) { *this = RotateXMatrix(sinAng,cosAng); }
	inline void RotateZ (type ang) { *this = RotateZMatrix(sinf(ang),cosf(ang)); }
	inline void RotateY (type ang) { *this = RotateYMatrix(sinf(ang),cosf(ang)); }
	inline void RotateX (type ang) { *this = RotateXMatrix(sinf(ang),cosf(ang)); }
T2	inline void Basis (const Vec3<type2> &c1, const Vec3<type2> &c2, const Vec3<type2> &c3) { *this = BasisMatrix(c1,c2,c3); }
};

// unary operators
MT	inline Matrix<type,rows,cols> operator - (const Matrix<type,rows,cols> &m) {return Matrix<type,rows,cols>::ZeroMatrix - m;}
MT	inline const Matrix<type,rows,cols>& operator + (const Matrix<type,rows,cols> &m) {return m;}
MT	inline Matrix<type,rows,cols>& operator + (Matrix<type,rows,cols> &m) {return m;}
/*
// disambiguate mat*mat from prim*mat
template <typename type, int rows, int cols, typename type2, int rows2, int cols2>
inline Matrix<type,rows,cols> operator * (const Matrix<type,rows,cols> m1, const Matrix<type2,rows2,cols2> &m2) {return m1.operator *(m2);}
*/
// prim * mat
template <typename type, int rows, int cols, typename type2>
inline Matrix<type,rows,cols> operator * (const type2 op, const Matrix<type,rows,cols> &m) {return m * op;}

//MT T2	inline Matrix<type,rows,cols> operator / (const type2 op, const Matrix<type,rows,cols> &m) {return m / op;}
//T2	inline VecT2(n) operator / (const float op, const VecT2(n) &v) {return v / op;}
//T2	inline VecT2(n) operator * (const int op, const VecT2(n) &v) {return v * op;}
//T2	inline VecT2(n) operator / (const int op, const VecT2(n) &v) {return v / op;}
//T2	inline const VecT2(n)&operator + (const VecT2(n) &v) {return v;}
//T2	inline VecT2(n)&operator + (VecT2(n) &v) {return v;}


// Implementation:
// (this should go in a .cpp file,
//  but it's a limitation of the language and the linker
//  that it would always have to be included anyway.)

MT void Matrix<type,rows,cols>::Transpose ()
{
	if(rows != cols) {assertMsg(0,"Can't transpose a non-square matrix in place (consider using GetTranspose() instead)."); return;}

	FOR(r,rows)
	FOR(c,r)
	{
		type temp = data[r][c];
		data[r][c] = data[c][r];
		data[c][r] = temp;
	}
}
MT Matrix<type,cols,rows> Matrix<type,rows,cols>::GetTranspose () const
{
	Matrix<type,cols,rows> m;
	FORALL(m[r][c] = data[c][r]);
	return m;
}

MT T2 void Matrix<type,rows,cols>::operator *= (const Matrix<type2,rows,cols>& op)
{
	Matrix<type,rows,cols> m;

	FOR(r,rows)
		FOR(c2,cols)
			FOR(r2,rows)
				m[r][c2] += data[r][r2] * op[r2][c2];

	(*this) = m;
}

MT MTM Matrix<type,rows,cols2> Matrix<type,rows,cols>::operator * (const Matrix<type2,cols,cols2>& op) const
{
	Matrix<type,rows,cols2> m;

	FOR(r,rows)
		FOR(c2,cols2)
			FOR(r2,rows)
				m[r][c2] += data[r][r2] * op[r2][c2];

	return m;
}




// values:
MT const Matrix<type,rows,cols> Matrix<type,rows,cols>::ZeroMatrix = Matrix<type,rows,cols>();

MT static Matrix<type,rows,cols> ComputeIdentityMatrix()
{
	Matrix<type,rows,cols> m;
	int size = rows < cols ? rows : cols;
	FOR(i, size)
		m[i][i] = (type)1;
	return m;
}
MT const Matrix<type,rows,cols> Matrix<type,rows,cols>::IdentityMatrix = ComputeIdentityMatrix<type,rows,cols>();

// Implementation:
// (this should go in a .cpp file,
//  but it's a limitation of the language and the linker
//  that it would always have to be included anyway.)

// These definitions were moved from gamex-vector.h, since
// they require the definition of Mat44f (and not meerly its
// declaration) to work...  All the more reason to be put in a
// .cpp file

// transforms a point by a matrix
T void Vec3<type>::operator*= (Mat44f& op)
{
	float* m = op.GetData();
	float xa, ya, za;
	xa  = x * (*m++);	ya  = x * (*m++);	za  = x * (*m++);	m++;
	xa += y * (*m++);	ya += y * (*m++);	za += y * (*m++);	m++;
	xa += z * (*m++);	ya += z * (*m++);	za += z * (*m++);	m++;
	xa +=     (*m++);	ya +=     (*m++);	za +=     (*m  );
	x = xa; y = ya; z = za;
}

// transforms a vector by a matrix (which is the same as transforming a point but without the translation)
T void Vec3<type>::TransformAsVectorBy (Mat44f& op)
{
	float* m = op.GetData();
	float xa, ya, za;
	xa  = x * (*m++);	ya  = x * (*m++);	za  = x * (*m++);	m++;
	xa += y * (*m++);	ya += y * (*m++);	za += y * (*m++);	m++;
	xa += z * (*m++);	ya += z * (*m++);	za += z * (*m++);
	x = xa; y = ya; z = za;
}

T void Vec4<type>::operator*= (Mat44f& op)
{
	float* m = op.GetData();
	float xa, ya, za, wa;
	xa  = x * (*m++);	ya  = x * (*m++);	za  = x * (*m++);	wa  = x * (*m++);
	xa += y * (*m++);	ya += y * (*m++);	za += y * (*m++);	wa += y * (*m++);
	xa += z * (*m++);	ya += z * (*m++);	za += z * (*m++);	wa += z * (*m++);
	xa += w * (*m++);	ya += w * (*m++);	za += w * (*m++);	wa += w * (*m  );
	x = xa; y = ya; z = za; w = wa;
}



#pragma pack(pop)


#undef LEN
#undef T
#undef T2
#undef MT
#undef MT2
#undef FORALL


// re-enable warnings
#pragma warning(pop)


#endif